# cython: language_level=3
# cython: c_api_binop_methods=True
# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from typing import Iterable, List, Sequence, TYPE_CHECKING, Tuple, Iterator
from libc.math cimport fabs, sin, cos, M_PI, hypot, atan2, acos, sqrt, fmod
import random

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL
    const double M_TAU

cdef double RAD2DEG = 180.0 / M_PI
cdef double DEG2RAD = M_PI / 180.0

if TYPE_CHECKING:
    from ezdxf.math import AnyVec, UVec

cdef bint isclose(double a, double b, double rel_tol, double abs_tol):
    # Has to match the Python implementation!
    cdef double diff = fabs(b - a)
    return diff <= fabs(rel_tol * b) or \
           diff <= fabs(rel_tol * a) or \
           diff <= abs_tol


cdef double normalize_rad_angle(double a):
    # Emulate the Python behavior of (a % math.tau)
    cdef double res = fmod(a, M_TAU)
    if res < 0.0:
        res += M_TAU
    return res

def _normalize_rad_angle(double angle):
    # just for testing
    return normalize_rad_angle(angle)

cdef double normalize_deg_angle(double a):
    # Emulate the Python behavior of (a % 360)
    cdef double res = fmod(a, 360.0)
    if res < 0.0:
        res += 360.0
    return res

def _normalize_deg_angle(double angle):
    # just for testing
    return normalize_deg_angle(angle)

cdef class Vec2:
    """ Immutable 2D vector.

    Init:

    - Vec2(vec2)
    - Vec2(vec3)
    - Vec2((x, y))
    - Vec2((x, y, z)), ignore z-axis
    - Vec2(x, y)
    - Vec2(x, y, z), ignore z-axis

    """

    def __cinit__(self, *args):
        cdef Py_ssize_t count = len(<tuple> args)

        if count == 0:  # fastest init - default constructor
            self.x = 0
            self.y = 0
            return

        if count == 1:
            arg = args[0]
            if isinstance(arg, Vec2):
                # fast init by Vec2()
                self.x = (<Vec2> arg).x
                self.y = (<Vec2> arg).y
                return
            if isinstance(arg, Vec3):
                # fast init by Vec3()
                self.x = (<Vec3> arg).x
                self.y = (<Vec3> arg).y
                return
            args = arg
            count = len(args)

        # ignore z-axis but raise error for 4 or more arguments
        if count > 3:
            raise TypeError('invalid argument count')

        # slow init by sequence
        self.x = args[0]
        self.y = args[1]

    def __reduce__(self):
        return Vec2, (self.x, self.y)

    @property
    def vec3(self) -> Vec3:
        return Vec3(self)

    def round(self, ndigits=None) -> Vec2:
        # only used for testing
        return Vec2(round(self.x, ndigits), round(self.y, ndigits))

    @staticmethod
    def list(items: Iterable[UVec]) -> List[Vec2]:
        return list(Vec2.generate(items))

    @staticmethod
    def tuple(items: Iterable[UVec]) -> Sequence[Vec2]:
        return tuple(Vec2.generate(items))

    @staticmethod
    def generate(items: Iterable[UVec]) -> Iterator[Vec2]:
        return (Vec2(item) for item in items)

    @staticmethod
    def from_angle(double angle, double length = 1.0) -> Vec2:
        return v2_from_angle(angle, length)

    @staticmethod
    def from_deg_angle(double angle, double length = 1.0) -> Vec2:
        return v2_from_angle(angle * DEG2RAD, length)

    def __str__(self) -> str:
        return f'({self.x}, {self.y})'

    def __repr__(self)-> str:
        return "Vec2" + self.__str__()

    def __len__(self) -> int:
        return 2

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def copy(self) -> Vec2:
        return self  # immutable

    def __copy__(self) -> Vec2:
        return self  # immutable

    def __deepcopy__(self, memodict: dict) -> Vec2:
        return self  # immutable

    def __getitem__(self, int index) -> float:
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        else:
            raise IndexError(f'invalid index {index}')

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y

    def __abs__(self) -> float:
        return hypot(self.x, self.y)

    @property
    def magnitude(self) -> float:
        return hypot(self.x, self.y)

    @property
    def is_null(self) -> bool:
        return fabs(self.x) <= ABS_TOL and fabs(self.y) <= ABS_TOL

    @property
    def angle(self) -> float:
        return atan2(self.y, self.x)

    @property
    def angle_deg(self) -> float:
        return atan2(self.y, self.x) * RAD2DEG

    def orthogonal(self, ccw: bool = True) -> Vec2:
        return v2_ortho(self, ccw)

    def lerp(self, other: "AnyVec", double factor = 0.5) -> Vec2:
        cdef Vec2 o = Vec2(other)
        return v2_lerp(self, o, factor)

    def normalize(self, double length = 1.) -> Vec2:
        return v2_normalize(self, length)

    def project(self, other: AnyVec) -> Vec2:
        cdef Vec2 o = Vec2(other)
        return v2_project(self, o)

    def __neg__(self) -> Vec2:
        cdef Vec2 res = Vec2()
        res.x = -self.x
        res.y = -self.y
        return res

    reversed = __neg__

    def __bool__(self) -> bool:
        return self.x != 0 or self.y != 0

    def isclose(self, other: UVec, *, double rel_tol=REL_TOL,
                double abs_tol = ABS_TOL) -> bool:
        cdef Vec2 o = Vec2(other)
        return isclose(self.x, o.x, rel_tol, abs_tol) and \
               isclose(self.y, o.y, rel_tol, abs_tol)

    def __eq__(self, other: UVec) -> bool:
        if not isinstance(other, Vec2):
            other = Vec2(other)
        return self.x == other.x and self.y == other.y

    def __lt__(self, other) -> bool:
        cdef Vec2 o = Vec2(other)
        if self.x == o.x:
            return self.y < o.y
        else:
            return self.x < o.x

    def __add__(self, other: AnyVec) -> Vec2:
        if not isinstance(other, Vec2):
            other = Vec2(other)
        return v2_add(self, <Vec2> other)

    # __radd__ not supported for Vec2

    __iadd__ = __add__  # immutable

    def __sub__(self, other: AnyVec) -> Vec2:
        if not isinstance(other, Vec2):
            other = Vec2(other)
        return v2_sub(self, <Vec2> other)

    # __rsub__ not supported for Vec2

    __isub__ = __sub__  # immutable

    def __mul__(self, factor) -> Vec2:
        if isinstance(self, Vec2):
            return v2_mul(self, factor)
        elif isinstance(factor, Vec2):
            return v2_mul(<Vec2> factor, <double> self)
        else:
            return NotImplemented

    # Special Cython <(3.0) feature: __rmul__ == __mul__(factor, self)

    def __rmul__(self, double factor) -> Vec2:
        # for Cython >= 3.0
        return v2_mul(self, factor)

    __imul__ = __mul__  # immutable

    def __truediv__(self, double factor) -> Vec2:
        return v2_mul(self, 1.0 / factor)

    # __rtruediv__ not supported -> TypeError

    def dot(self, other: AnyVec) -> float:
        cdef Vec2 o = Vec2(other)
        return v2_dot(self, o)

    def det(self, other: AnyVec) -> float:
        cdef Vec2 o = Vec2(other)
        return v2_det(self, o)

    def distance(self, other: AnyVec) -> float:
        cdef Vec2 o = Vec2(other)
        return v2_dist(self, o)

    def angle_between(self, other: AnyVec) -> float:
        cdef Vec2 o = Vec2(other)
        return v2_angle_between(self, o)

    def rotate(self, double angle) -> Vec2:
        cdef double self_angle = atan2(self.y, self.x)
        cdef double magnitude = hypot(self.x, self.y)
        return v2_from_angle(self_angle + angle, magnitude)

    def rotate_deg(self, double angle) -> Vec2:
        return self.rotate(angle * DEG2RAD)

    @staticmethod
    def sum(items: Iterable[Vec2]) -> Vec2:
        cdef Vec2 res = Vec2()
        cdef Vec2 tmp
        res.x = 0.0
        res.y = 0.0
        for v in items:
            tmp = v
            res.x += tmp.x
            res.y += tmp.y
        return res


cdef Vec2 v2_add(Vec2 a, Vec2 b):
    res = Vec2()
    res.x = a.x + b.x
    res.y = a.y + b.y
    return res

cdef Vec2 v2_sub(Vec2 a, Vec2 b):
    res = Vec2()
    res.x = a.x - b.x
    res.y = a.y - b.y
    return res

cdef Vec2 v2_mul(Vec2 a, double factor):
    res = Vec2()
    res.x = a.x * factor
    res.y = a.y * factor
    return res

cdef double v2_dot(Vec2 a, Vec2 b):
    return a.x * b.x + a.y * b.y

cdef double v2_det(Vec2 a, Vec2 b):
    return a.x * b.y - a.y * b.x

cdef double v2_dist(Vec2 a, Vec2 b):
    return hypot(a.x - b.x, a.y - b.y)

cdef Vec2 v2_from_angle(double angle, double length):
    cdef Vec2 res = Vec2()
    res.x = cos(angle) * length
    res.y = sin(angle) * length
    return res

cdef double v2_angle_between(Vec2 a, Vec2 b) except -1000:
    cdef double cos_theta = v2_dot(v2_normalize(a, 1.0), v2_normalize(b, 1.0))
    # avoid domain errors caused by floating point imprecision:
    if cos_theta < -1.0:
        cos_theta = -1.0
    elif cos_theta > 1.0:
        cos_theta = 1.0
    return acos(cos_theta)

cdef Vec2 v2_normalize(Vec2 a, double length):
    cdef double factor = length / hypot(a.x, a.y)
    cdef Vec2 res = Vec2()
    res.x = a.x * factor
    res.y = a.y * factor
    return res

cdef Vec2 v2_lerp(Vec2 a, Vec2 b, double factor):
    cdef Vec2 res = Vec2()
    res.x = a.x + (b.x - a.x) * factor
    res.y = a.y + (b.y - a.y) * factor
    return res

cdef Vec2 v2_ortho(Vec2 a, bint ccw):
    cdef Vec2 res = Vec2()
    if ccw:
        res.x = -a.y
        res.y = a.x
    else:
        res.x = a.y
        res.y = -a.x
    return res

cdef Vec2 v2_project(Vec2 a, Vec2 b):
    cdef Vec2 uv = v2_normalize(a, 1.0)
    return v2_mul(uv, v2_dot(uv, b))

cdef bint v2_isclose(Vec2 a, Vec2 b, double rel_tol, double abs_tol):
    return isclose(a.x, b.x, rel_tol, abs_tol) and \
           isclose(a.y, b.y, rel_tol, abs_tol)


cdef class Vec3:
    """ Immutable 3D vector.

    Init:

    - Vec3()
    - Vec3(vec3)
    - Vec3(vec2)
    - Vec3((x, y))
    - Vec3((x, y, z))
    - Vec3(x, y)
    - Vec3(x, y, z)

    """

    def __cinit__(self, *args):
        cdef Py_ssize_t count = len(<tuple> args)
        if count == 0:  # fastest init - default constructor
            self.x = 0
            self.y = 0
            self.z = 0
            return

        if count == 1:
            arg0 = args[0]
            if isinstance(arg0, Vec3):
                # fast init by Vec3()
                self.x = (<Vec3> arg0).x
                self.y = (<Vec3> arg0).y
                self.z = (<Vec3> arg0).z
                return
            if isinstance(arg0, Vec2):
                # fast init by Vec2()
                self.x = (<Vec2> arg0).x
                self.y = (<Vec2> arg0).y
                self.z = 0
                return
            args = arg0
            count = len(args)

        if count > 3 or count < 2:
            raise TypeError('invalid argument count')

        # slow init by sequence
        self.x = args[0]
        self.y = args[1]
        if count > 2:
            self.z = args[2]
        else:
            self.z = 0.0

    def __reduce__(self):
        return Vec3, self.xyz

    @property
    def xy(self) -> Vec3:
        cdef Vec3 res = Vec3()
        res.x = self.x
        res.y = self.y
        return res

    @property
    def xyz(self) -> Tuple[float, float, float]:
        return self.x, self.y, self.z

    @property
    def vec2(self) -> Vec2:
        cdef Vec2 res = Vec2()
        res.x = self.x
        res.y = self.y
        return res

    def replace(self, x: float = None, y: float = None,
                z: float = None) -> Vec3:
        cdef Vec3 res = Vec3()
        res.x = self.x if x is None else x
        res.y = self.y if y is None else y
        res.z = self.z if z is None else z
        return res

    def round(self, ndigits: int | None = None) -> Vec3:
        return Vec3(
            round(self.x, ndigits),
            round(self.y, ndigits),
            round(self.z, ndigits),
        )

    @staticmethod
    def list(items: Iterable[UVec]) -> List[Vec3]:
        return list(Vec3.generate(items))

    @staticmethod
    def tuple(items: Iterable[UVec]) -> Sequence[Vec3]:
        return tuple(Vec3.generate(items))

    @staticmethod
    def generate(items: Iterable[UVec]) -> Iterator[Vec3]:
        return (Vec3(item) for item in items)

    @staticmethod
    def from_angle(double angle, double length = 1.0) -> Vec3:
        return v3_from_angle(angle, length)

    @staticmethod
    def from_deg_angle(double angle, double length = 1.0) -> Vec3:
        return v3_from_angle(angle * DEG2RAD, length)

    @staticmethod
    def random(double length = 1.0) -> Vec3:
        cdef Vec3 res = Vec3()
        uniform = random.uniform
        res.x = uniform(-1, 1)
        res.y = uniform(-1, 1)
        res.z = uniform(-1, 1)
        return v3_normalize(res, length)

    def __str__(self) -> str:
        return f'({self.x}, {self.y}, {self.z})'

    def __repr__(self)-> str:
        return "Vec3" + self.__str__()

    def __len__(self) -> int:
        return 3

    def __hash__(self) -> int:
        return hash(self.xyz)

    def copy(self) -> Vec3:
        return self  # immutable

    __copy__ = copy

    def __deepcopy__(self, memodict: dict) -> Vec3:
        return self  # immutable!

    def __getitem__(self, int index) -> float:
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise IndexError(f'invalid index {index}')

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y
        yield self.z

    def __abs__(self) -> float:
        return v3_magnitude(self)

    @property
    def magnitude(self) -> float:
        return v3_magnitude(self)

    @property
    def magnitude_xy(self) -> float:
        return hypot(self.x, self.y)

    @property
    def magnitude_square(self) -> float:
        return v3_magnitude_sqr(self)

    @property
    def is_null(self) -> bool:
        return fabs(self.x) <= ABS_TOL and fabs(self.y) <= ABS_TOL and \
               fabs(self.z) <= ABS_TOL

    def is_parallel(self, other: UVec, *, double rel_tol=REL_TOL,
                    double abs_tol = ABS_TOL) -> bool:
        cdef Vec3 o = Vec3(other)
        cdef Vec3 v1 = v3_normalize(self, 1.0)
        cdef Vec3 v2 = v3_normalize(o, 1.0)
        cdef Vec3 neg_v2 = v3_reverse(v2)
        return v3_isclose(v1, v2, rel_tol, abs_tol) or \
               v3_isclose(v1, neg_v2, rel_tol, abs_tol)

    @property
    def spatial_angle(self) -> float:
        return acos(v3_dot(<Vec3> X_AXIS, v3_normalize(self, 1.0)))

    @property
    def spatial_angle_deg(self) -> float:
        return self.spatial_angle * RAD2DEG

    @property
    def angle(self) -> float:
        return atan2(self.y, self.x)

    @property
    def angle_deg(self) -> float:
        return atan2(self.y, self.x) * RAD2DEG

    def orthogonal(self, ccw: bool = True) -> Vec3:
        return v3_ortho(self, ccw)

    def lerp(self, other: UVec, double factor=0.5) -> Vec3:
        if not isinstance(other, Vec3):
            other = Vec3(other)
        return v3_lerp(self, <Vec3> other, factor)

    def project(self, other: UVec) -> Vec3:
        if not isinstance(other, Vec3):
            other = Vec3(other)
        return v3_project(self, <Vec3> other)

    def normalize(self, double length = 1.) -> Vec3:
        return v3_normalize(self, length)

    def reversed(self) -> Vec3:
        return v3_reverse(self)

    def __neg__(self) -> Vec3:
        return v3_reverse(self)

    def __bool__(self) -> bool:
        return not self.is_null

    def isclose(self, other: UVec, *, double rel_tol = REL_TOL,
                double abs_tol = ABS_TOL) -> bool:
        if not isinstance(other, Vec3):
            other = Vec3(other)
        return v3_isclose(self, <Vec3> other, rel_tol, abs_tol)

    def __eq__(self, other: UVec) -> bool:
        if not isinstance(other, Vec3):
            other = Vec3(other)
        return self.x == other.x and self.y == other.y and self.z == other.z

    def __lt__(self, other: UVec) -> bool:
        if not isinstance(other, Vec3):
            other = Vec3(other)
        if self.x == (<Vec3> other).x:
            return self.y < (<Vec3> other).y
        else:
            return self.x < (<Vec3> other).x

    # Special Cython (<3.0) feature: __radd__ == __add__(other, self)
    def __add__(self, other) -> Vec3:
        if not isinstance(self, Vec3):
            # other is the real self
            return v3_add(Vec3(self), <Vec3> other)

        if not isinstance(other, Vec3):
            other = Vec3(other)
        return v3_add(<Vec3> self, <Vec3> other)

    __radd__ = __add__  # Cython >= 3.0

    __iadd__ = __add__  # immutable

    # Special Cython (<3.0) feature: __rsub__ == __sub__(other, self)
    def __sub__(self, other) -> Vec3:
        if not isinstance(self, Vec3):
            # other is the real self
            return v3_sub(Vec3(self), <Vec3> other)

        if not isinstance(other, Vec3):
            other = Vec3(other)
        return v3_sub(<Vec3> self, <Vec3> other)

    def __rsub__(self, other) -> Vec3:
        # for Cython >= 3.0
        return v3_sub(Vec3(other), <Vec3> self)

    __isub__ = __sub__  # immutable

    # Special Cython <(3.0) feature: __rmul__ == __mul__(factor, self)
    def __mul__(self, factor) -> Vec3:
        if isinstance(factor, Vec3):
            return v3_mul(<Vec3> factor, self)
        return v3_mul(<Vec3> self, factor)

    def __rmul__(self, double factor) -> Vec3:
        # for Cython >= 3.0
        return v3_mul(self, factor)

    __imul__ = __mul__  # immutable

    def __truediv__(self, double factor) -> Vec3:
        return v3_mul(self, 1.0 / factor)

    # __rtruediv__ not supported -> TypeError

    @staticmethod
    def sum(items: Iterable[UVec]) -> Vec3:
        cdef Vec3 res = Vec3()
        cdef Vec3 tmp
        for v in items:
            tmp = Vec3(v)
            res.x += tmp.x
            res.y += tmp.y
            res.z += tmp.z
        return res

    def dot(self, other: UVec) -> float:
        cdef Vec3 o = Vec3(other)
        return v3_dot(self, o)

    def cross(self, other: UVec) -> Vec3:
        cdef Vec3 o = Vec3(other)
        return v3_cross(self, o)

    def distance(self, other: UVec) -> float:
        cdef Vec3 o = Vec3(other)
        return v3_dist(self, o)

    def angle_between(self, other: UVec) -> float:
        cdef Vec3 o = Vec3(other)
        return v3_angle_between(self, o)

    def angle_about(self, base: UVec, target: UVec) -> float:
        cdef Vec3 b = Vec3(base)
        cdef Vec3 t = Vec3(target)
        return v3_angle_about(self, b, t)

    def rotate(self, double angle) -> Vec3:
        cdef double angle_ = atan2(self.y, self.x) + angle
        cdef double magnitude_ = hypot(self.x, self.y)
        cdef Vec3 res = Vec3.from_angle(angle_, magnitude_)
        res.z = self.z
        return res

    def rotate_deg(self, double angle) -> Vec3:
        return self.rotate(angle * DEG2RAD)


X_AXIS = Vec3(1, 0, 0)
Y_AXIS = Vec3(0, 1, 0)
Z_AXIS = Vec3(0, 0, 1)
NULLVEC = Vec3(0, 0, 0)

cdef Vec3 v3_add(Vec3 a, Vec3 b):
    res = Vec3()
    res.x = a.x + b.x
    res.y = a.y + b.y
    res.z = a.z + b.z
    return res

cdef Vec3 v3_sub(Vec3 a, Vec3 b):
    res = Vec3()
    res.x = a.x - b.x
    res.y = a.y - b.y
    res.z = a.z - b.z
    return res


cdef Vec3 v3_mul(Vec3 a, double factor):
    res = Vec3()
    res.x = a.x * factor
    res.y = a.y * factor
    res.z = a.z * factor
    return res


cdef Vec3 v3_reverse(Vec3 a):
    cdef Vec3 res = Vec3()
    res.x = -a.x
    res.y = -a.y
    res.z = -a.z
    return res


cdef double v3_dot(Vec3 a, Vec3 b):
    return a.x * b.x + a.y * b.y + a.z * b.z


cdef Vec3 v3_cross(Vec3 a, Vec3 b):
    res = Vec3()
    res.x = a.y * b.z - a.z * b.y
    res.y = a.z * b.x - a.x * b.z
    res.z = a.x * b.y - a.y * b.x
    return res


cdef inline double v3_magnitude_sqr(Vec3 a):
    return a.x * a.x + a.y * a.y + a.z * a.z


cdef inline double v3_magnitude(Vec3 a):
    return sqrt(v3_magnitude_sqr(a))


cdef double v3_dist(Vec3 a, Vec3 b):
    cdef double dx = a.x - b.x
    cdef double dy = a.y - b.y
    cdef double dz = a.z - b.z
    return sqrt(dx * dx + dy * dy + dz * dz)


cdef Vec3 v3_from_angle(double angle, double length):
    cdef Vec3 res = Vec3()
    res.x = cos(angle) * length
    res.y = sin(angle) * length
    return res


cdef double v3_angle_between(Vec3 a, Vec3 b) except -1000:
    cdef double cos_theta = v3_dot(v3_normalize(a, 1.0), v3_normalize(b, 1.0))
    # avoid domain errors caused by floating point imprecision:
    if cos_theta < -1.0:
        cos_theta = -1.0
    elif cos_theta > 1.0:
        cos_theta = 1.0
    return acos(cos_theta)


cdef double v3_angle_about(Vec3 a, Vec3 base, Vec3 target):
    cdef Vec3 x_axis = v3_normalize(v3_sub(base, v3_project(a, base)), 1.0)
    cdef Vec3 y_axis = v3_normalize(v3_cross(a, x_axis), 1.0)
    cdef double target_projected_x = v3_dot(x_axis, target)
    cdef double target_projected_y = v3_dot(y_axis, target)
    return normalize_rad_angle(atan2(target_projected_y, target_projected_x))


cdef Vec3 v3_normalize(Vec3 a, double length):
    cdef double factor = length / v3_magnitude(a)
    cdef Vec3 res = Vec3()
    res.x = a.x * factor
    res.y = a.y * factor
    res.z = a.z * factor
    return res


cdef Vec3 v3_lerp(Vec3 a, Vec3 b, double factor):
    cdef Vec3 res = Vec3()
    res.x = a.x + (b.x - a.x) * factor
    res.y = a.y + (b.y - a.y) * factor
    res.z = a.z + (b.z - a.z) * factor
    return res


cdef Vec3 v3_ortho(Vec3 a, bint ccw):
    cdef Vec3 res = Vec3()
    res.z = a.z
    if ccw:
        res.x = -a.y
        res.y = a.x
    else:
        res.x = a.y
        res.y = -a.x
    return res


cdef Vec3 v3_project(Vec3 a, Vec3 b):
    cdef Vec3 uv = v3_normalize(a, 1.0)
    return v3_mul(uv, v3_dot(uv, b))


cdef bint v3_isclose(Vec3 a, Vec3 b, double rel_tol, double abs_tol):
    return isclose(a.x, b.x, rel_tol, abs_tol) and \
           isclose(a.y, b.y, rel_tol, abs_tol) and \
           isclose(a.z, b.z, rel_tol, abs_tol)


def distance(p1: UVec, p2: UVec) -> float:
    cdef Vec3 a = Vec3(p1)
    cdef Vec3 b = Vec3(p2)
    return v3_dist(a, b)


def lerp(p1: UVec, p2: UVec, double factor = 0.5) -> Vec3:
    cdef Vec3 a = Vec3(p1)
    cdef Vec3 b = Vec3(p2)
    return v3_lerp(a, b, factor)



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module", params=["R12", "R2000"])
def dxf(request, tmpdir_factory):
    doc = ezdxf.new()
    msp = doc.modelspace()
    msp.add_line((0, 0), (1, 0))
    psp = doc.layout()
    psp.add_circle((0, 0), 1)
    filename = tmpdir_factory.mktemp(request.param).join("test.dxf")
    doc.dxfversion = request.param
    doc.saveas(filename)
    return filename


def test_load_dxf(dxf):
    doc = ezdxf.readfile(dxf)

    msp = doc.modelspace()
    assert len(msp) == 1
    assert msp[0].dxftype() == "LINE"

    psp = doc.layout()
    assert len(psp) == 1
    assert psp[0].dxftype() == "CIRCLE"



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
import os
import ezdxf

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"
COLDFIRE = os.path.join(
    ezdxf.EZDXF_TEST_FILES, "CADKitSamples/kit-dev-coldfire-xilinx_5213.dxf"
)


@pytest.mark.skipif(
    not os.path.exists(COLDFIRE), reason="test data not present"
)
def test_kit_dev_coldfire():
    doc = ezdxf.readfile(COLDFIRE)
    auditor = doc.audit()
    assert len(auditor) == 0


@pytest.fixture(params=["Leica_Disto_S910.dxf"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip(f"File {filename} not found.")
    return filename


def test_leica_disto_r12(filename):
    doc = ezdxf.readfile(filename)
    auditor = doc.audit()
    assert len(auditor) == 0



# Copyright (c) 2011-2024, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.lldxf import const
from ezdxf.entities import MText, Insert


@pytest.fixture
def doc():
    return ezdxf.new()


@pytest.fixture
def block(doc):
    block = doc.blocks.new("TEST")
    block.add_attdef("TAG1", (0, 0), text="default text1")
    block.add_attdef("TAG2", (0, 5), text="default text2")
    return block


def test_auto_blockref(doc, block):
    values = {"TAG1": "text1", "TAG2": "text2"}
    msp = doc.modelspace()
    blockref = msp.add_auto_blockref("TEST", (0, 0), values)
    autoblock = doc.blocks[blockref.dxf.name]
    entities = list(autoblock)
    insert = entities[0]
    assert isinstance(insert, Insert) is True

    attrib1, attrib2 = insert.attribs
    assert attrib1.dxftype() == "ATTRIB"
    assert attrib1.dxf.tag == "TAG1"
    assert attrib1.dxf.text == "text1"
    assert attrib1.dxf.attribute_type == const.ATTRIB_TYPE_SINGLE_LINE

    assert attrib2.dxftype() == "ATTRIB"
    assert attrib2.dxf.tag == "TAG2"
    assert attrib2.dxf.text == "text2"
    assert attrib2.dxf.attribute_type == const.ATTRIB_TYPE_SINGLE_LINE


def test_add_auto_attribs(doc, block):
    values = {"TAG1": "text1", "TAG2": "text2"}
    msp = doc.modelspace()
    blockref = msp.add_blockref("TEST", (0, 0)).add_auto_attribs(values)
    assert blockref.dxftype() == "INSERT"

    attrib1, attrib2 = blockref.attribs
    assert attrib1.dxftype() == "ATTRIB"
    assert attrib1.dxf.tag == "TAG1"
    assert attrib1.dxf.text == "text1"
    assert attrib1.dxf.attribute_type == const.ATTRIB_TYPE_SINGLE_LINE

    assert attrib2.dxftype() == "ATTRIB"
    assert attrib2.dxf.tag == "TAG2"
    assert attrib2.dxf.text == "text2"
    assert attrib2.dxf.attribute_type == const.ATTRIB_TYPE_SINGLE_LINE


def test_add_auto_attribs_with_default_text(doc, block):
    msp = doc.modelspace()
    blockref = msp.add_blockref("TEST", (0, 0)).add_auto_attribs({})
    attrib1, attrib2 = blockref.attribs
    assert attrib1.dxf.text == "default text1"
    assert attrib2.dxf.text == "default text2"


def test_add_auto_attribs_with_missing_tag():
    doc = ezdxf.new()
    msp = doc.modelspace()
    blocks = doc.blocks.new("TEST")
    attdef = blocks.add_attdef("TAG1", (0, 0))
    attdef.dxf.discard("tag")  # simulate erroneous DXF input

    blockref = msp.add_blockref("TEST", (0, 0)).add_auto_attribs({"TAG1": "Test"})
    assert blockref.has_attrib("TAG1") is False


def test_add_auto_attribs_with_missing_location():
    doc = ezdxf.new()
    msp = doc.modelspace()
    blocks = doc.blocks.new("TEST")
    attdef = blocks.add_attdef("TAG1", (0, 0))
    attdef.dxf.discard("insert")  # simulate erroneous DXF input

    blockref = msp.add_blockref("TEST", (0, 0)).add_auto_attribs({"TAG1": "Test"})
    assert blockref.has_attrib("TAG1") is False


def test_has_attdef(block):
    assert block.has_attdef("TAG1") is True
    assert block.has_attdef("TAG_Z") is False


def test_get_attdef(block):
    attdef = block.get_attdef("TAG1")
    assert attdef.dxf.tag == "TAG1"
    assert block.get_attdef("TAG1_Z") is None


def test_get_attdef_text(block):
    block.add_attdef("TAGX", insert=(0, 0), text="PRESET_TEXT")
    text = block.get_attdef_text("TAGX")
    assert text == "PRESET_TEXT"


def test_attdef_getter_properties(block):
    attrib = block.get_attdef("TAG1")

    assert attrib.is_const is False
    assert attrib.is_invisible is False
    assert attrib.is_verify is False
    assert attrib.is_preset is False


def test_attdef_setter_properties(block):
    attrib = block.get_attdef("TAG1")

    assert attrib.is_const is False
    attrib.is_const = True
    assert attrib.is_const is True

    assert attrib.is_invisible is False
    attrib.is_invisible = True
    assert attrib.is_invisible is True

    assert attrib.is_verify is False
    attrib.is_verify = True
    assert attrib.is_verify is True

    assert attrib.is_preset is False
    attrib.is_preset = True
    assert attrib.is_preset is True


class TestMultiLineAttribs:
    @pytest.fixture
    def doc(self):
        return ezdxf.new()

    @pytest.fixture
    def block(self, doc):
        block = doc.blocks.new("TEST")
        attdef = block.add_attdef("TAG1")
        # Note: \n is converted to \P automatically at DXF export
        mtext = MText.new(dxfattribs={"insert": (1, 1), "text": "default\ntext"})
        attdef.embed_mtext(mtext)
        return block

    def test_has_multi_line_attdef(self, block):
        attdef = block.get_attdef("TAG1")
        assert attdef is not None
        assert attdef.has_embedded_mtext_entity is True
        assert attdef.dxf.attribute_type == const.ATTDEF_TYPE_MULTI_LINE

    def test_multi_line_attdef_properties(self, block):
        attdef = block.get_attdef("TAG1")
        assert attdef is not None
        assert attdef.dxf.insert == (1, 1)
        assert attdef.dxf.text == "default", "expected just first line"

        mtext = attdef.virtual_mtext_entity()
        assert isinstance(mtext, MText) is True

        assert mtext.dxf.insert == (1, 1)
        assert mtext.dxf.text == "default\ntext", "expected all lines"

    def test_add_auto_multi_line_attrib(self, doc, block):
        msp = doc.modelspace()
        insert = msp.add_blockref(block.name, insert=(2, 3))
        assert insert is not None

        insert.add_auto_attribs({"TAG1": "MultiLine\nAttribute"})

        attrib = insert.attribs[0]
        assert attrib is not None

        assert attrib.dxf.insert == (3, 4)
        assert attrib.dxf.text == "MultiLine", "expected just first line"
        assert attrib.has_embedded_mtext_entity is True
        assert attrib.dxf.attribute_type == const.ATTRIB_TYPE_MULTI_LINE

        mtext = attrib.virtual_mtext_entity()
        assert isinstance(mtext, MText) is True

        assert mtext.dxf.insert == (3, 4)
        assert mtext.dxf.text == "MultiLine\nAttribute", "expected all lines"



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.entities import Insert, Body
from ezdxf.layouts import Modelspace
from ezdxf.math import Matrix44, Vec3
from ezdxf.protocols import SupportsTemporaryTransformation


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new("R2007")
    return doc.modelspace()


@pytest.fixture()
def body(msp: Modelspace):
    return msp.add_body()


def test_supports_temporary_transform(body: Body):
    assert isinstance(body, SupportsTemporaryTransformation) is True


def test_add_and_get_transform(body: Body):
    m = Matrix44.translate(10, 20, 30)
    tt = body.temporary_transformation()
    tt.set_matrix(m)

    tt2 = body.temporary_transformation()
    m2 = tt2.get_matrix()
    assert isinstance(m2, Matrix44)

    v = Vec3(1, 1, 1)
    assert m.transform(v).isclose(m2.transform(v))


def test_apply_temp_transform(body: Body, msp: Modelspace):
    m = Matrix44.translate(10, 20, 30)
    tt = body.temporary_transformation()
    tt.set_matrix(m)

    done = tt.apply_transformation(body)
    assert done is True

    # BODY was replaced by a block reference to an anonymous block
    insert = msp[-1]
    assert isinstance(insert, Insert)
    assert insert.dxf.name.startswith("*U")
    assert insert.dxf.insert.isclose((10, 20, 30))

    # anonymous block contains BODY
    block = insert.block()
    assert block is not None
    assert block[0] is body


def test_transform_and_copy(body: Body):
    m = Matrix44.translate(10, 20, 30)
    tt = body.temporary_transformation()
    tt.set_matrix(m)

    copy = body.copy()
    tt2 = copy.temporary_transformation()
    m2 = tt2.get_matrix()
    assert isinstance(m2, Matrix44)

    v = Vec3(1, 1, 1)
    assert m.transform(v).isclose(m2.transform(v))


def test_copy_has_independent_temporary_transformation(msp: Modelspace):
    """Copy and original must not share the same TransformByBlockReference.

    Regression test: Body.copy_data() previously did a shallow assignment of
    _temporary_transformation.  Transforming the copy (e.g. via
    bbox.extents() -> virtual_entities()) would contaminate the original with
    a spurious pending transform, destroying block structure on save.
    """
    body = msp.add_body()
    copy = body.copy()

    # Must be separate objects
    assert body.temporary_transformation() is not copy.temporary_transformation()

    # Transforming the copy must not affect the original
    copy.transform(Matrix44.translate(10, 20, 30))
    assert body.temporary_transformation().get_matrix() is None
    assert copy.temporary_transformation().get_matrix() is not None


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020-2022 Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.math import intersection_line_line_2d, Vec2


def vec2(x, y):
    return Vec2((x, y))


def test_intersect_virtual():
    ray1 = (vec2(10, 1), vec2(20, 10))
    ray2 = (vec2(17, -7), vec2(-10, 3))

    point = intersection_line_line_2d(ray1, ray2)
    assert point.isclose(vec2(5.7434, -2.8309), abs_tol=1e-4)


def test_intersect_with_vertical():
    ray1 = (vec2(10, 1), vec2(10, -7))
    ray2 = (vec2(-10, 3), vec2(17, -7))
    point = intersection_line_line_2d(ray1, ray2)
    assert point.x == 10
    assert point.isclose(vec2(10.0, -4.4074), abs_tol=1e-4)


def test_intersect_with_horizontal():
    ray1 = (vec2(-10, 10), vec2(10, 10))
    ray2 = (vec2(-10, 20), vec2(10, 0))
    point = intersection_line_line_2d(ray1, ray2)
    assert point.y == 10
    assert point.isclose(vec2(0.0, 10.0), abs_tol=1e-4)


def test_intersect_with_vertical_and_horizontal():
    ray1 = (vec2(-10, 10), vec2(10, 10))
    ray2 = (vec2(5, 0), vec2(5, 20))
    point = intersection_line_line_2d(ray1, ray2)
    assert point.y == 10
    assert point.x == 5
    assert point.isclose(vec2(5.0, 10.0), abs_tol=1e-4)


def test_intersect_parallel_vertical():
    ray1 = (vec2(10, 1), vec2(10, -7))
    ray2 = (vec2(12, -10), vec2(12, 7))
    assert intersection_line_line_2d(ray1, ray2) is None


def test_intersect_parallel_horizontal():
    ray3 = (vec2(11, 0), vec2(-11, 0))
    ray4 = (vec2(0, 0), vec2(1, 0))
    assert intersection_line_line_2d(ray3, ray4) is None


def test_intersect_normal_vertical():
    ray = (vec2(10, 1), vec2(10, -7))
    ortho = (vec2(0, 3), vec2(10, 3))
    point = intersection_line_line_2d(ray, ortho)
    assert point.isclose(vec2(10, 3))


def test_intersection_by_line_end_points():
    hor = (vec2(1, 1), vec2(3, 1))
    vert = (vec2(2, 1), vec2(2, 5))
    point = intersection_line_line_2d(hor, vert)
    assert point.isclose(vec2(2, 1))

    vert = (vec2(2, 5), vec2(2, 1))
    point = intersection_line_line_2d(hor, vert)
    assert point.isclose(vec2(2, 1))


def test_intersect_real():
    line1 = (vec2(0, 0), vec2(4, 4))
    line2 = (vec2(3, 2), vec2(5, 0))
    point = intersection_line_line_2d(line1, line2, virtual=False)
    assert point is None


def test_intersect_real_colinear():
    line1 = (vec2(0, 0), vec2(4, 4))
    line2 = (vec2(2, 2), vec2(4, 0))  # intersection point, is endpoint of ray2
    point = intersection_line_line_2d(line1, line2, virtual=False)
    assert point.isclose(vec2(2, 2))


@pytest.mark.parametrize(
    "p2", [(4, 0), (0, 4), (4, 4)], ids=["horiz", "vert", "diag"]
)
def test_intersect_coincident_lines(p2):
    line1 = (Vec2(0, 0), Vec2(p2))
    point = intersection_line_line_2d(line1, line1, virtual=False)
    assert point is None


def test_issue_128():
    line1 = (vec2(175.0, 5.0), vec2(175.0, 50.0))
    line2 = (vec2(-10.1231, 30.1235), vec2(300.2344, 30.1235))
    point = intersection_line_line_2d(line1, line2, virtual=False)
    assert point is not None
    assert point.isclose(vec2(175.0, 30.1235))


def test_issue_664():
    from ezdxf.math import ConstructionLine
    p1 = Vec2(16399619.87946683, -199438.8133075837)
    p2 = Vec2(16399700.34999472, -199438.8133075837)
    p3 = Vec2(16399659.76235549, -199423.8133075837)
    p4 = Vec2(16399659.76235549, -199463.8133075837)
    expected = Vec2(16399659.76235549, -199438.8133075837)

    point = intersection_line_line_2d(
        (p1, p2), (p3, p4), virtual=False
    )
    assert point.isclose(expected)

    cl1 = ConstructionLine(p1, p2)
    cl2 = ConstructionLine(p3, p4)
    assert cl1.intersect(cl2).isclose(expected)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2016-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Optional,
    TYPE_CHECKING,
    Sequence,
    Any,
    Iterator,
)
from functools import partial
import logging
from .tags import DXFTag
from .types import POINT_CODES, NONE_TAG, VALID_XDATA_GROUP_CODES

if TYPE_CHECKING:
    from ezdxf.eztypes import Tags

logger = logging.getLogger("ezdxf")


def tag_reorder_layer(tagger: Iterable[DXFTag]) -> Iterator[DXFTag]:
    """Reorder coordinates of legacy DXF Entities, for now only LINE.

    Input Raw tag filter.

    Args:
        tagger: low level tagger

    """

    collector: Optional[list] = None
    for tag in tagger:
        if tag.code == 0:
            if collector is not None:
                # stop collecting if inside a supported entity
                entity = _s(collector[0].value)
                yield from COORDINATE_FIXING_TOOLBOX[entity](collector)  # type: ignore
                collector = None

            if _s(tag.value) in COORDINATE_FIXING_TOOLBOX:
                collector = [tag]
                # do not yield collected tag yet
                tag = NONE_TAG
        else:  # tag.code != 0
            if collector is not None:
                collector.append(tag)
                # do not yield collected tag yet
                tag = NONE_TAG
        if tag is not NONE_TAG:
            yield tag


# invalid point codes if not part of a point started with 1010, 1011, 1012, 1013
INVALID_Y_CODES = {code + 10 for code in POINT_CODES}
INVALID_Z_CODES = {code + 20 for code in POINT_CODES}
# A single group code 38 is an elevation tag (e.g. LWPOLYLINE)
# Is (18, 28, 38?) is a valid point code?
INVALID_Z_CODES.remove(38)
INVALID_CODES = INVALID_Y_CODES | INVALID_Z_CODES
X_CODES = POINT_CODES


def filter_invalid_point_codes(tagger: Iterable[DXFTag]) -> Iterable[DXFTag]:
    """Filter invalid and misplaced point group codes.

    - removes x-axis without following y-axis
    - removes y- and z-axis without leading x-axis

    Args:
        tagger: low level tagger

    """

    def entity() -> str:
        if handle_tag:
            return f"in entity #{_s(handle_tag[1])}"
        else:
            return ""

    expected_code = -1
    z_code = 0
    point: list[Any] = []
    handle_tag = None
    for tag in tagger:
        code = tag[0]
        if code == 5:  # ignore DIMSTYLE entity
            handle_tag = tag
        if point and code != expected_code:
            # at least x, y axis is required else ignore point
            if len(point) > 1:
                yield from point
            else:
                logger.info(
                    f"remove misplaced x-axis tag: {str(point[0])}" + entity()
                )
            point.clear()

        if code in X_CODES:
            expected_code = code + 10
            z_code = code + 20
            point.append(tag)
        elif code == expected_code:
            point.append(tag)
            expected_code += 10
            if expected_code > z_code:
                expected_code = -1
        else:
            # ignore point group codes without leading x-axis
            if code not in INVALID_CODES:
                yield tag
            else:
                axis = "y-axis" if code in INVALID_Y_CODES else "z-axis"
                logger.info(
                    f"remove misplaced {axis} tag: {str(tag)}" + entity()
                )

    if len(point) == 1:
        logger.info(f"remove misplaced x-axis tag: {str(point[0])}" + entity())
    elif len(point) > 1:
        yield from point


def fix_coordinate_order(tags: Tags, codes: Sequence[int] = (10, 11)):
    def extend_codes():
        for code in codes:
            yield code  # x tag
            yield code + 10  # y tag
            yield code + 20  # z tag

    def get_coords(code: int):
        # if x or y coordinate is missing, it is a DXFStructureError
        # but here is not the location to validate the DXF structure
        try:
            yield coordinates[code]
        except KeyError:
            pass
        try:
            yield coordinates[code + 10]
        except KeyError:
            pass
        try:
            yield coordinates[code + 20]
        except KeyError:
            pass

    coordinate_codes = frozenset(extend_codes())
    coordinates = {}
    remaining_tags = []
    insert_pos = None
    for tag in tags:
        # separate tags
        if tag.code in coordinate_codes:
            coordinates[tag.code] = tag
            if insert_pos is None:
                insert_pos = tags.index(tag)
        else:
            remaining_tags.append(tag)

    if len(coordinates) == 0:
        # no coordinates found, this is probably a DXFStructureError,
        # but here is not the location to validate the DXF structure,
        # just do nothing.
        return tags

    ordered_coords = []
    for code in codes:
        ordered_coords.extend(get_coords(code))
    remaining_tags[insert_pos:insert_pos] = ordered_coords
    return remaining_tags


COORDINATE_FIXING_TOOLBOX = {
    "LINE": partial(fix_coordinate_order, codes=(10, 11)),
}


def filter_invalid_xdata_group_codes(
    tags: Iterable[DXFTag],
) -> Iterator[DXFTag]:
    return (tag for tag in tags if tag.code in VALID_XDATA_GROUP_CODES)


def filter_invalid_handles(tags: Iterable[DXFTag]) -> Iterator[DXFTag]:
    line = -1
    handle_code = 5
    structure_tag = ""
    for tag in tags:
        line += 2
        if tag.code == 0:
            structure_tag = tag.value
            if _s(tag.value) == "DIMSTYLE":
                handle_code = 105
            else:
                handle_code = 5
        elif tag.code == handle_code:
            try:
                int(tag.value, 16)
            except ValueError:
                logger.warning(
                    f'skipped invalid handle "{_s(tag.value)}" in '
                    f'DXF entity "{_s(structure_tag)}" near line {line}'
                )
                continue
        yield tag


def _s(b) -> str:
    if isinstance(b, bytes):
        return b.decode(encoding="ascii", errors="ignore")
    return b



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.math.shape import Shape2d


@pytest.fixture()
def square():
    return Shape2d([(0, 0), (1, 0), (1, 1), (0, 1)])


def test_init(square):

    # shape is list like
    assert len(square) == 4

    # elements are Vec3s?
    assert square[0].x == 0
    assert square[0].y == 0
    assert square[2].x == 1
    assert square[2].y == 1

    e = list(square)
    assert len(e) == 4

    square.append((3, 3))
    assert square[-1] == (3, 3)


def test_translate(square):
    square.translate((1, 0))
    assert square[0] == (1, 0)
    assert square[2] == (2, 1)


def test_scale(square):
    square.scale(sx=2, sy=3)
    assert square[0] == (0, 0)
    assert square[2] == (2, 3)


def test_scale_uniform(square):
    square.scale_uniform(1.5)
    assert square[0] == (0, 0)
    assert square[2] == (1.5, 1.5)


def test_rotate(square):
    square.rotate(90)
    assert square[0].isclose((0, 0))
    assert square[2].isclose((-1, 1))
    square.rotate(90)
    assert square[0].isclose((0, 0))
    assert square[2].isclose((-1, -1))


def test_rotate_center(square):
    square.translate((2, 2))
    square.rotate(90, center=(2, 2))
    assert square[0].isclose((2, 2))
    assert square[2].isclose((1, 3))



# Copyright (c) 2019-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Optional, Iterator, Sequence, TypeVar, Generic
import abc
import math
import numpy as np

from ezdxf.math import Vec3, Vec2, UVec

T = TypeVar("T", Vec2, Vec3)

__all__ = ["BoundingBox2d", "BoundingBox", "AbstractBoundingBox"]


class AbstractBoundingBox(Generic[T], abc.ABC):
    extmin: T
    extmax: T

    @abc.abstractmethod
    def __init__(self, vertices: Optional[Iterable[UVec]] = None):
        ...

    def copy(self):
        box = self.__class__()
        box.extmin = self.extmin
        box.extmax = self.extmax
        return box

    def __str__(self) -> str:
        return f"[{self.extmin}, {self.extmax}]"

    def __repr__(self) -> str:
        name = self.__class__.__name__
        if self.has_data:
            return f"{name}({self.__str__()})"
        else:
            return f"{name}()"

    def __iter__(self) -> Iterator[T]:
        if self.has_data:
            yield self.extmin
            yield self.extmax

    @abc.abstractmethod
    def extend(self, vertices: Iterable[UVec]) -> None:
        ...

    @property
    @abc.abstractmethod
    def is_empty(self) -> bool:
        ...

    @abc.abstractmethod
    def inside(self, vertex: UVec) -> bool:
        ...

    @abc.abstractmethod
    def has_intersection(self, other: AbstractBoundingBox[T]) -> bool:
        ...

    @abc.abstractmethod
    def has_overlap(self, other: AbstractBoundingBox[T]) -> bool:
        ...

    @abc.abstractmethod
    def intersection(self, other: AbstractBoundingBox[T]) -> AbstractBoundingBox[T]:
        ...

    def contains(self, other: AbstractBoundingBox[T]) -> bool:
        """Returns ``True`` if the `other` bounding box is completely inside
        this bounding box.

        """
        return self.inside(other.extmin) and self.inside(other.extmax)

    def any_inside(self, vertices: Iterable[UVec]) -> bool:
        """Returns ``True`` if any vertex is inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.has_data:
            return any(self.inside(v) for v in vertices)
        return False

    def all_inside(self, vertices: Iterable[UVec]) -> bool:
        """Returns ``True`` if all vertices are inside this bounding box.

        Vertices at the box border are inside!
        """
        if self.has_data:
            # all() returns True for an empty set of vertices
            has_any = False
            for v in vertices:
                has_any = True
                if not self.inside(v):
                    return False
            return has_any
        return False

    @property
    def has_data(self) -> bool:
        """Returns ``True`` if the bonding box has known limits."""
        return math.isfinite(self.extmin.x)

    @property
    def size(self) -> T:
        """Returns size of bounding box."""
        return self.extmax - self.extmin

    @property
    def center(self) -> T:
        """Returns center of bounding box."""
        return self.extmin.lerp(self.extmax)

    def union(self, other: AbstractBoundingBox[T]) -> AbstractBoundingBox[T]:
        """Returns a new bounding box as union of this and `other` bounding
        box.
        """
        vertices: list[T] = []
        if self.has_data:
            vertices.extend(self)
        if other.has_data:
            vertices.extend(other)
        return self.__class__(vertices)

    def rect_vertices(self) -> Sequence[Vec2]:
        """Returns the corners of the bounding box in the xy-plane as
        :class:`Vec2` objects.
        """
        if self.has_data:
            x0, y0, *_ = self.extmin
            x1, y1, *_ = self.extmax
            return Vec2(x0, y0), Vec2(x1, y0), Vec2(x1, y1), Vec2(x0, y1)
        else:
            raise ValueError("empty bounding box")

    def grow(self, value: float) -> None:
        """Grow or shrink the bounding box by an uniform value in x, y and
        z-axis. A negative value shrinks the bounding box.
        Raises :class:`ValueError` for shrinking the size of the bounding box to
        zero or below in any dimension.
        """
        if self.has_data:
            if value < 0.0:
                min_ext = min(self.size)
                if -value >= min_ext / 2.0:
                    raise ValueError("shrinking one or more dimensions <= 0")
            self.extmax += Vec3(value, value, value)
            self.extmin += Vec3(-value, -value, -value)


class BoundingBox(AbstractBoundingBox[Vec3]):
    """3D bounding box.

    Args:
        vertices: iterable of ``(x, y, z)`` tuples or :class:`Vec3` objects

    """

    __slots__ = ("extmin", "extmax")

    def __init__(self, vertices: Optional[Iterable[UVec]] = None):
        self.extmin = Vec3(math.inf, math.inf, math.inf)
        self.extmax = self.extmin
        if vertices is not None:
            try:
                self.extmin, self.extmax = extents3d(vertices)
            except ValueError:
                # No or invalid data creates an empty BoundingBox
                pass

    @property
    def is_empty(self) -> bool:
        """Returns ``True`` if the bounding box is empty or the bounding box
        has a size of 0 in any or all dimensions or is undefined.

        """
        if self.has_data:
            sx, sy, sz = self.size
            return sx * sy * sz == 0.0
        return True

    def extend(self, vertices: Iterable[UVec]) -> None:
        """Extend bounds by `vertices`.

        Args:
            vertices: iterable of vertices

        """
        v = list(vertices)
        if not v:
            return
        if self.has_data:
            v.extend([self.extmin, self.extmax])
        self.extmin, self.extmax = extents3d(v)

    def inside(self, vertex: UVec) -> bool:
        """Returns ``True`` if `vertex` is inside this bounding box.

        Vertices at the box border are inside!
        """
        if not self.has_data:
            return False
        x, y, z = Vec3(vertex).xyz
        xmin, ymin, zmin = self.extmin.xyz
        xmax, ymax, zmax = self.extmax.xyz
        return (xmin <= x <= xmax) and (ymin <= y <= ymax) and (zmin <= z <= zmax)

    def has_intersection(self, other: AbstractBoundingBox[T]) -> bool:
        """Returns ``True`` if this bounding box intersects with `other` but does
        not include touching bounding boxes, see also :meth:`has_overlap`::

            bbox1 = BoundingBox([(0, 0, 0), (1, 1, 1)])
            bbox2 = BoundingBox([(1, 1, 1), (2, 2, 2)])
            assert bbox1.has_intersection(bbox2) is False

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        # Check for a separating axis:
        if not self.has_data or not other.has_data:
            return False
        o_min = Vec3(other.extmin)  # could be a 2D bounding box
        o_max = Vec3(other.extmax)  # could be a 2D bounding box

        # Check for a separating axis:
        if self.extmin.x >= o_max.x:
            return False
        if self.extmax.x <= o_min.x:
            return False
        if self.extmin.y >= o_max.y:
            return False
        if self.extmax.y <= o_min.y:
            return False
        if self.extmin.z >= o_max.z:
            return False
        if self.extmax.z <= o_min.z:
            return False
        return True

    def has_overlap(self, other: AbstractBoundingBox[T]) -> bool:
        """Returns ``True`` if this bounding box intersects with `other` but
        in contrast to :meth:`has_intersection` includes touching bounding boxes too::

            bbox1 = BoundingBox([(0, 0, 0), (1, 1, 1)])
            bbox2 = BoundingBox([(1, 1, 1), (2, 2, 2)])
            assert bbox1.has_overlap(bbox2) is True

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        # Check for a separating axis:
        if not self.has_data or not other.has_data:
            return False
        o_min = Vec3(other.extmin)  # could be a 2D bounding box
        o_max = Vec3(other.extmax)  # could be a 2D bounding box
        # Check for a separating axis:
        if self.extmin.x > o_max.x:
            return False
        if self.extmax.x < o_min.x:
            return False
        if self.extmin.y > o_max.y:
            return False
        if self.extmax.y < o_min.y:
            return False
        if self.extmin.z > o_max.z:
            return False
        if self.extmax.z < o_min.z:
            return False
        return True

    def cube_vertices(self) -> Sequence[Vec3]:
        """Returns the 3D corners of the bounding box as :class:`Vec3` objects."""
        if self.has_data:
            x0, y0, z0 = self.extmin
            x1, y1, z1 = self.extmax
            return (
                Vec3(x0, y0, z0),
                Vec3(x1, y0, z0),
                Vec3(x1, y1, z0),
                Vec3(x0, y1, z0),
                Vec3(x0, y0, z1),
                Vec3(x1, y0, z1),
                Vec3(x1, y1, z1),
                Vec3(x0, y1, z1),
            )
        else:
            raise ValueError("empty bounding box")

    def intersection(self, other: AbstractBoundingBox[T]) -> BoundingBox:
        """Returns the bounding box of the intersection cube of both
        3D bounding boxes. Returns an empty bounding box if the intersection
        volume is 0.

        """
        new_bbox = self.__class__()
        if not self.has_intersection(other):
            return new_bbox
        s_min_x, s_min_y, s_min_z = Vec3(self.extmin)
        o_min_x, o_min_y, o_min_z = Vec3(other.extmin)
        s_max_x, s_max_y, s_max_z = Vec3(self.extmax)
        o_max_x, o_max_y, o_max_z = Vec3(other.extmax)
        new_bbox.extend(
            [
                (
                    max(s_min_x, o_min_x),
                    max(s_min_y, o_min_y),
                    max(s_min_z, o_min_z),
                ),
                (
                    min(s_max_x, o_max_x),
                    min(s_max_y, o_max_y),
                    min(s_max_z, o_max_z),
                ),
            ]
        )
        return new_bbox


class BoundingBox2d(AbstractBoundingBox[Vec2]):
    """2D bounding box.

    Args:
        vertices: iterable of ``(x, y[, z])`` tuples or :class:`Vec3` objects

    """

    __slots__ = ("extmin", "extmax")

    def __init__(self, vertices: Optional[Iterable[UVec]] = None):
        self.extmin = Vec2(math.inf, math.inf)
        self.extmax = self.extmin
        if vertices is not None:
            try:
                self.extmin, self.extmax = extents2d(vertices)
            except ValueError:
                # No or invalid data creates an empty BoundingBox
                pass

    @property
    def is_empty(self) -> bool:
        """Returns ``True`` if the bounding box is empty. The bounding box has a
        size of 0 in any or all dimensions or is undefined.
        """
        if self.has_data:
            sx, sy = self.size
            return sx * sy == 0.0
        return True

    def extend(self, vertices: Iterable[UVec]) -> None:
        """Extend bounds by `vertices`.

        Args:
            vertices: iterable of vertices

        """
        v = list(vertices)
        if not v:
            return
        if self.has_data:
            v.extend([self.extmin, self.extmax])
        self.extmin, self.extmax = extents2d(v)

    def inside(self, vertex: UVec) -> bool:
        """Returns ``True`` if `vertex` is inside this bounding box.

        Vertices at the box border are inside!
        """
        if not self.has_data:
            return False
        v = Vec2(vertex)
        min_ = self.extmin
        max_ = self.extmax
        return (min_.x <= v.x <= max_.x) and (min_.y <= v.y <= max_.y)

    def has_intersection(self, other: AbstractBoundingBox[T]) -> bool:
        """Returns ``True`` if this bounding box intersects with `other` but does
        not include touching bounding boxes, see also :meth:`has_overlap`::

            bbox1 = BoundingBox2d([(0, 0), (1, 1)])
            bbox2 = BoundingBox2d([(1, 1), (2, 2)])
            assert bbox1.has_intersection(bbox2) is False

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        if not self.has_data or not other.has_data:
            return False
        # Check for a separating axis:
        if self.extmin.x >= other.extmax.x:
            return False
        if self.extmax.x <= other.extmin.x:
            return False
        if self.extmin.y >= other.extmax.y:
            return False
        if self.extmax.y <= other.extmin.y:
            return False
        return True

    def intersection(self, other: AbstractBoundingBox[T]) -> BoundingBox2d:
        """Returns the bounding box of the intersection rectangle of both
        2D bounding boxes. Returns an empty bounding box if the intersection
        area is 0.
        """
        new_bbox = self.__class__()
        if not self.has_intersection(other):
            return new_bbox
        s_min_x, s_min_y = Vec2(self.extmin)
        o_min_x, o_min_y = Vec2(other.extmin)
        s_max_x, s_max_y = Vec2(self.extmax)
        o_max_x, o_max_y = Vec2(other.extmax)
        new_bbox.extend(
            [
                (max(s_min_x, o_min_x), max(s_min_y, o_min_y)),
                (min(s_max_x, o_max_x), min(s_max_y, o_max_y)),
            ]
        )
        return new_bbox

    def has_overlap(self, other: AbstractBoundingBox[T]) -> bool:
        """Returns ``True`` if this bounding box intersects with `other` but
        in contrast to :meth:`has_intersection` includes touching bounding boxes too::

            bbox1 = BoundingBox2d([(0, 0), (1, 1)])
            bbox2 = BoundingBox2d([(1, 1), (2, 2)])
            assert bbox1.has_overlap(bbox2) is True

        """
        # Source: https://gamemath.com/book/geomtests.html#intersection_two_aabbs
        if not self.has_data or not other.has_data:
            return False
        # Check for a separating axis:
        if self.extmin.x > other.extmax.x:
            return False
        if self.extmax.x < other.extmin.x:
            return False
        if self.extmin.y > other.extmax.y:
            return False
        if self.extmax.y < other.extmin.y:
            return False
        return True


def extents3d(vertices: Iterable[UVec]) -> tuple[Vec3, Vec3]:
    """Returns the extents of the bounding box as tuple (extmin, extmax)."""
    vertices = np.array([Vec3(v).xyz for v in vertices], dtype=np.float64)
    if len(vertices):
        return Vec3(vertices.min(0)), Vec3(vertices.max(0))
    else:
        raise ValueError("no vertices given")


def extents2d(vertices: Iterable[UVec]) -> tuple[Vec2, Vec2]:
    """Returns the extents of the bounding box as tuple (extmin, extmax)."""
    vertices = np.array([(x, y) for x, y, *_ in vertices], dtype=np.float64)
    if len(vertices):
        return Vec2(vertices.min(0)), Vec2(vertices.max(0))
    else:
        raise ValueError("no vertices given")



# Copyright (c) 2013-2022, Manfred Moitzi
# License: MIT-License
import pytest
import ezdxf

from ezdxf.query import EntityQuery, name_query
from ezdxf.entities import Text, Line, Circle, Arc, MText
from ezdxf.math import Vec3
from ezdxf import colors


class TestNameQuery:
    @pytest.fixture
    def entities(self):
        return ["LINE", "CIRCLE", "POLYLINE", "MESH", "SOLID"]

    def test_all_names(self, entities):
        result = list(name_query(entities, "*"))
        assert result == entities

    def test_some_names(self, entities):
        result = list(name_query(entities, "LINE SOLID"))
        assert result == ["LINE", "SOLID"]

    def test_exclude_some_names(self, entities):
        result = list(name_query(entities, "* !SOLID"))
        assert "SOLID" not in result


def test_remove_supports_virtual_entities():
    result = EntityQuery([Text(), Line(), Arc()]).remove("TEXT")
    assert len(result) == 2


@pytest.fixture(scope="module")
def modelspace():
    doc = ezdxf.new()
    modelspace = doc.modelspace()
    modelspace.add_line((0, 0), (10, 0), {"layer": "lay_lines", "color": 7})
    modelspace.add_polyline2d(
        [(0, 0), (3, 1), (7, 4), (10, 0)],
        dxfattribs={"layer": "lay_lines", "color": 6},
    )
    modelspace.add_text("TEST", dxfattribs={"layer": "lay_text", "color": 6})
    modelspace.add_circle((0, 0), 1, dxfattribs={"layer": "π"})
    # just 4 entities: LINE, TEXT, POLYLINE, CIRCLE
    # VERTEX & SEQEND are linked to the POLYLINE entity and do not appear in any entity space
    return doc.modelspace()


def test_empty_init():
    result = EntityQuery()
    assert len(result) == 0


def test_select_all(modelspace):
    result = EntityQuery(modelspace, "*")
    # 1xLINE, 1xPOLYLINE, 0xVERTEX, 0xSEQEND, 1x TEXT, 1x CIRCLE
    assert len(result.entities) == 4
    assert len(result) == 4


def test_first(modelspace):
    result = EntityQuery(modelspace, "*")
    assert result.first.dxftype() == "LINE"


def test_last(modelspace):
    result = EntityQuery(modelspace, "*")
    assert result.last.dxftype() == "CIRCLE"


def test_new_query_select_all(modelspace):
    result = ezdxf.query.new(modelspace, "*")
    # 1xLINE, 1xPOLYLINE, 0xVERTEX, 0xSEQEND, 1x TEXT, 1x CIRCLE
    assert len(result.entities) == 4
    assert len(result) == 4


def test_new_empty_query():
    result = ezdxf.query.new()
    assert len(result.entities) == 0
    assert len(result) == 0


def test_select_line(modelspace):
    result = EntityQuery(modelspace, "LINE")
    assert len(result) == 1, "should be 1 LINE"


def test_select_layer_1(modelspace):
    result = EntityQuery(modelspace, '*[layer=="lay_lines"]')
    assert len(result) == 2
    assert result.first.dxftype() == "LINE"
    assert result.last.dxftype() == "POLYLINE"


def test_select_unicode_layer_name(modelspace):
    result = EntityQuery(modelspace, '*[layer=="π"]')
    assert len(result) == 1
    assert result.first.dxftype() == "CIRCLE"


def test_select_layer_1_exclude_line(modelspace):
    result = EntityQuery(modelspace, '* !LINE[layer=="lay_lines"]')
    # 1xPOLYLINE
    assert len(result) == 1


def test_match_regex(modelspace):
    result = EntityQuery(modelspace, '*[layer ? "lay_.*"]')
    assert len(result) == 3


def test_match_regex_not_text(modelspace):
    result = EntityQuery(modelspace, '* !TEXT[layer ? "lay_.*"]')
    assert len(result) == 2


def test_match_whole_string(modelspace):
    # re just compares the start of a string, check for an
    # implicit '$' at the end of the search string.
    result = EntityQuery(modelspace, '*[layer=="lay"]')
    assert len(result) == 0


def test_not_supported_attribute(modelspace):
    result = EntityQuery(modelspace, '*[mozman!="TEST"]')
    assert len(result) == 0


def test_bool_select(modelspace):
    result = EntityQuery(modelspace, '*[layer=="lay_lines" & color==7]')
    # 1xLINE
    assert len(result) == 1


def test_bool_select_2(modelspace):
    result = EntityQuery(
        modelspace, '*[layer=="lay_lines" & color==7 | color==6]'
    )
    # 1xLINE(layer=="lay_lines" & color==7) 1xPOLYLINE(color==6) 1xTEXT(color==6)
    assert len(result) == 3


def test_bool_select_3(modelspace):
    result = EntityQuery(
        modelspace, '*[layer=="lay_lines" & (color==7 | color==6)]'
    )
    # 1xLINE(layer=="lay_lines" & color==7) 1xPOLYLINE(layer=="lay_lines" & color==6)
    assert len(result) == 2


def test_bool_select_4(modelspace):
    result = EntityQuery(
        modelspace, '*[(layer=="lay_lines" | layer=="lay_text") & !color==7]'
    )
    # 1xPOLYLINE(layer=="lay_lines" & color==6) 1xTEXT(layer=="lay_text" & color==6)
    assert len(result) == 2


def test_ignore_case(modelspace):
    result = EntityQuery(modelspace, '*[layer=="LAY_lines"]i')
    # 1xLINE 1xPOLYLINE
    assert len(result) == 2


def test_ignore_case_for_num_values(modelspace):
    result = EntityQuery(modelspace, "*[color==6]i")
    # 1xPOLYLINE 1xTEXT
    assert len(result) == 2


def test_ignore_case_match_regex(modelspace):
    result = EntityQuery(modelspace, '*[layer ? "LaY_.*"]i')
    assert len(result) == 3


def test_extend_query(modelspace):
    result = EntityQuery(modelspace, "*")
    length = len(result)
    result.extend(result)
    assert len(result) == length


def test_all_names():
    names = "ONE TWO THREE"
    result = " ".join(name_query(names.split(), "*"))
    assert names == result


def test_match_one_string():
    names = "ONE TWO THREE"
    result = list(name_query(names.split(), "ONE"))
    assert "ONE" == result[0]


def test_match_full_string():
    names = "ONEONE TWO THREE"
    result = list(name_query(names.split(), "ONE"))
    assert len(result) == 0


def test_match_default_values():
    """An attribute query should match the default value if the attribute is
    not present.
    """
    text = Text()
    assert text.dxf.hasattr("style") is False
    result = EntityQuery([text], "*[style=='Standard']")
    assert result.first is text


def test_iter_does_not_remove_destroyed_entities():
    query = EntityQuery([Text(), Text(), Text()])
    query[0].destroy()
    assert len(query) == 3
    assert len(list(query)) == 3


def test_purge_does_remove_destroyed_entities():
    query = EntityQuery([Text(), Text(), Text()])
    query[0].destroy()
    query.purge()
    assert len(query) == 2


class TestGetSetDelItemInterface:
    def test_select_entities_with_supported_attribute(self, modelspace):
        """EntityQuery.__getitem__("attribute") returns all entities which
        support the given DXF attribute
        """
        query = modelspace.query()
        assert len(query) == 4, "expected all entities"
        lines = query["start"]
        assert len(lines) == 1
        assert lines.selected_dxf_attribute == "start"

        assert lines.first.dxftype() == "LINE"
        assert query["center"].first.dxftype() == "CIRCLE"

    def test_set_item_accepts_only_strings_as_key(self):
        query = EntityQuery()
        with pytest.raises(TypeError):
            query[1] = "has to fail"
        with pytest.raises(TypeError):
            query[:] = "has to fail"

    def test_set_item_set_supported_dxf_attributes(self):
        query = EntityQuery([Line(), Circle(), Text()])
        query["layer"] = "MyLayer"
        assert all(e.dxf.layer == "MyLayer" for e in query) is True

    def test_set_item_ignores_unsupported_attributes(self):
        query = EntityQuery([Line(), Text()])
        query["text"] = "MyText"
        assert query.query("TEXT").first.dxf.text == "MyText"

    def test_set_item_does_not_ignore_invalid_attribute_values(self):
        query = EntityQuery([Line(), Text()])
        with pytest.raises(TypeError):
            query["start"] = "InvalidVertex"
        with pytest.raises(ValueError):
            query["color"] = "red"

    def test_del_item_ignores_unsupported_attributes(self):
        query = EntityQuery([Line(), Text()])
        query["layer"] = "MyLayer"
        del query["layer"]
        assert query[0].dxf.layer == "0", "expected the default value"
        assert query[1].dxf.layer == "0", "expected the default value"


class TestVirtualAttributes:
    @pytest.fixture
    def entities(self):
        return EntityQuery([
            MText.new(dxfattribs={"text": "content"}),  # virtual text attribute
            Text.new(dxfattribs={"text": "content"})
        ])

    def test_get_virtual_attributes(self, entities):
        assert len(entities["text"]) == 2

    def test_set_virtual_attributes(self, entities):
        entities["text"] = "XYZ"
        assert all(e.dxf.text == "XYZ" for e in entities)

    def test_cannot_delete_virtual_attributes(self, entities):
        """Cannot delete virtual DXF attributes! """
        del entities["text"]
        assert entities.query("MTEXT").first.dxf.text == "content"


class TestOperatorSelection:
    @pytest.fixture
    def query(self):
        return EntityQuery(
            [
                Line.new(dxfattribs={"layer": "Lay1", "color": 1}),
                Text.new(dxfattribs={"layer": "Lay2", "color": 2}),
            ]
        )

    def test_no_selected_dxf_attribute_raises_type_error(self):
        query = EntityQuery()
        with pytest.raises(TypeError):
            _ = query == "MyLayer"

    def test_is_equal_operator(self, query: EntityQuery):
        result = query["layer"] == "Lay1"
        assert len(result) == 1
        assert result.first.dxftype() == "LINE"
        assert (
            result.selected_dxf_attribute == "layer"
        ), "selection should be propagated"

    def test_is_not_equal_operator(self, query: EntityQuery):
        result = query["layer"] != "Lay1"
        assert len(result) == 1
        assert result.first.dxftype() == "TEXT"

    def test_query_is_case_insensitive_by_default(self, query: EntityQuery):
        result = query["layer"] == "LAY1"
        assert len(result) == 1

    def test_less_than_operator(self, query: EntityQuery):
        result = query["color"] < 2
        assert len(result) == 1
        assert result.first.dxftype() == "LINE"

    def test_less_than_raises_type_error_for_vector_attributes(self):
        query = EntityQuery([Text.new(dxfattribs={"insert": Vec3(0, 0)})])
        with pytest.raises(TypeError):
            _ = query["insert"] < (1, 0)

    def test_greater_than_operator(self, query: EntityQuery):
        result = query["color"] > 1
        assert len(result) == 1
        assert result.first.dxftype() == "TEXT"

    def test_greater_than_raises_type_error_for_vector_attributes(self):
        query = EntityQuery([Text.new(dxfattribs={"insert": Vec3(0, 0)})])
        with pytest.raises(TypeError):
            _ = query["insert"] > (1, 0)

    def test_less_equal_operator(self, query: EntityQuery):
        result = query["color"] <= 2
        assert len(result) == 2

    def test_less_equal_raises_type_error_for_vector_attributes(self):
        query = EntityQuery([Text.new(dxfattribs={"insert": Vec3(0, 0)})])
        with pytest.raises(TypeError):
            _ = query["insert"] <= (1, 0)

    def test_greater_equal_operator(self, query: EntityQuery):
        result = query["color"] >= 1
        assert len(result) == 2

    def test_greater_equal_raises_type_error_for_vector_attributes(self):
        query = EntityQuery([Text.new(dxfattribs={"insert": Vec3(0, 0)})])
        with pytest.raises(TypeError):
            _ = query["insert"] >= (1, 0)

    def test_filter_operator(self, query):
        """Build your own operator to filter non DXF attributes."""
        for e in query:
            e.rgb = (0, 0, 0)
        result = query.filter(lambda e: e.rgb == (0, 0, 0))
        assert len(result) == 2


class TestRegexMatch:
    @pytest.fixture
    def base(self):
        return EntityQuery(
            [
                Line.new(dxfattribs={"layer": "Lay_line", "color": 1}),
                Circle.new(
                    dxfattribs={"layer": "lAy_circle", "center": Vec3(0, 0)}
                ),
                Text.new(dxfattribs={"layer": "laY_text"}),
            ]
        )

    def test_match_nothing(self, base: EntityQuery):
        result = base["layer"].match("nothing")
        assert len(result) == 0

    def test_match_is_case_insensitive(self, base: EntityQuery):
        result = base["layer"].match("lay_.*")
        assert len(result) == 3

    def test_match_case_insensitive(self, base: EntityQuery):
        layers = base["layer"]
        layers.ignore_case = False
        assert len(layers.match("Lay_.*")) == 1

    def test_match_invalid_attributes_raise_type_error(self, base: EntityQuery):
        with pytest.raises(TypeError):
            base["color"].match("nothing")
        with pytest.raises(TypeError):
            base["center"].match("nothing")


class TestSetOperations:
    @pytest.fixture
    def base(self):
        return EntityQuery(
            [
                Line.new(dxfattribs={"layer": "line", "color": 1}),
                Circle.new(dxfattribs={"layer": "circle", "color": 2}),
                Text.new(dxfattribs={"layer": "text", "color": 3}),
            ]
        )

    def test_union(self, base: EntityQuery):
        other = EntityQuery(
            [
                Line.new(dxfattribs={"layer": "line", "color": 1}),
                Circle.new(dxfattribs={"layer": "circle", "color": 2}),
                Text.new(dxfattribs={"layer": "text", "color": 3}),
            ]
        )
        assert len(base | other) == 6

    def test_union_unique_entities(self, base: EntityQuery):
        assert len(base | base) == 3

    def test_difference(self, base: EntityQuery):
        result = base - base.query("LINE")
        assert len(result) == 2
        assert set([e.dxftype() for e in result]) == {"CIRCLE", "TEXT"}

    def test_intersection(self, base: EntityQuery):
        result = base & base.query("LINE")
        assert len(result) == 1
        assert [e.dxftype() for e in result] == ["LINE"]

    def test_symmetric_difference(self, base: EntityQuery):
        other = base.query("LINE") | EntityQuery(
            [Arc.new(dxfattribs={"layer": "arc", "color": 4})]
        )
        result = base ^ other
        assert len(result) == 3
        assert set(e.dxftype() for e in result) == {"CIRCLE", "TEXT", "ARC"}

    def test_operator_combination(self, base: EntityQuery):
        result = (base["layer"] == "line") | (base["color"] == 2)
        assert len(result) == 2
        assert set(e.dxftype() for e in result) == {"LINE", "CIRCLE"}


class TestDescriptors:
    @pytest.fixture
    def entities(self):
        return EntityQuery(
            [
                Line.new(
                    handle="ABBA",
                    dxfattribs={
                        "layer": "line",
                        "color": 1,
                        "linetype": "SOLID",
                        "lineweight": 18,
                    },
                ),
                Circle.new(
                    handle="FEFE",
                    dxfattribs={
                        "layer": "circle",
                        "color": 2,
                        "linetype": "DOT",
                        "lineweight": 25,
                    },
                ),
            ]
        )

    def test_select_layers(self, entities: EntityQuery):
        assert len(entities.layer == "LINE") == 1

    def test_set_layer_attribute(self, entities: EntityQuery):
        entities.layer = "TEST"
        assert len(entities) == 2
        assert all(e.dxf.layer == "TEST" for e in entities) is True

    def test_delete_layer_attribute(self, entities: EntityQuery):
        del entities.layer
        assert len(entities) == 2
        assert all(e.dxf.layer == "0" for e in entities) is True

    def test_color_descriptor_exists(self, entities: EntityQuery):
        assert len(entities.color == 1) == 1
        # set and delete follow the same schema as for layer

    def test_linetype_descriptor_exists(self, entities: EntityQuery):
        assert len(entities.linetype == "DOT") == 1
        # set and delete follow the same schema as for layer

    def test_lineweight_descriptor_exists(self, entities: EntityQuery):
        assert len(entities.lineweight == 18) == 1
        # set and delete follow the same schema as for layer

    def test_ltscale_descriptor_exists(self, entities: EntityQuery):
        assert len(entities.ltscale == 1.0) == 2, "default value"
        # set and delete follow the same schema as for layer

    def test_invisible_descriptor_exists(self, entities: EntityQuery):
        assert len(entities.invisible == 0) == 2, "default value"
        # set and delete follow the same schema as for layer

    def test_true_color_descriptor_exists(self, entities: EntityQuery):
        assert (
            len(entities.true_color == colors.rgb2int((0, 0, 0))) == 0
        ), "has no default value"
        # set and delete follow the same schema as for layer

    def test_transparency_descriptor_exists(self, entities: EntityQuery):
        assert (
            len(entities.transparency == colors.float2transparency(0)) == 0
        ), "has no default value"
        # set and delete follow the same schema as for layer



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
import math
from ezdxf.audit import AuditError
from ezdxf.lldxf import validator
from ezdxf.math import (
    Vec3,
    Matrix44,
    NULLVEC,
    X_AXIS,
    Z_AXIS,
    ellipse,
    ConstructionEllipse,
    OCS,
)
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import (
    DXFGraphic,
    acdb_entity,
    add_entity,
    replace_entity,
    acdb_entity_group_codes,
)
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entities import DXFNamespace, Spline
    from ezdxf.audit import Auditor


__all__ = ["Ellipse"]

MIN_RATIO = 1e-10  # tested with DWG TrueView 2022
MAX_RATIO = 1.0  # tested with DWG TrueView 2022
TOL = 1e-9  # arbitrary choice


def is_valid_ratio(ratio: float) -> bool:
    """Check if axis-ratio is in valid range, takes an upper bound tolerance into
    account for floating point imprecision.
    """
    return MIN_RATIO <= abs(ratio) < (MAX_RATIO + TOL)


def clamp_axis_ratio(ratio: float) -> float:
    """Clamp axis-ratio into valid range and remove possible floating point imprecision.
    """
    sign = -1 if ratio < 0 else +1
    ratio = abs(ratio)
    if ratio < MIN_RATIO:
        return MIN_RATIO * sign
    if ratio > MAX_RATIO:
        return MAX_RATIO * sign
    return ratio * sign


acdb_ellipse = DefSubclass(
    "AcDbEllipse",
    {
        "center": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Major axis vector from 'center':
        "major_axis": DXFAttr(
            11,
            xtype=XType.point3d,
            default=X_AXIS,
            validator=validator.is_not_null_vector,
        ),
        # The extrusion vector does not establish an OCS, it is just the normal
        # vector of the ellipse plane:
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Ratio has to be in the range: -1.0 ... -1e-10 and +1e-10 ... +1.0:
        "ratio": DXFAttr(
            40, default=MAX_RATIO, validator=is_valid_ratio, fixer=clamp_axis_ratio
        ),
        # Start of ellipse, this value is 0.0 for a full ellipse:
        "start_param": DXFAttr(41, default=0),
        # End of ellipse, this value is 2π for a full ellipse:
        "end_param": DXFAttr(42, default=math.tau),
    },
)
acdb_ellipse_group_code = group_code_mapping(acdb_ellipse)
merged_ellipse_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_ellipse_group_code  # type: ignore
)

HALF_PI = math.pi / 2.0


@register_entity
class Ellipse(DXFGraphic):
    """DXF ELLIPSE entity"""

    DXFTYPE = "ELLIPSE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_ellipse)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        # bypass DXFGraphic, loading proxy graphic is skipped!
        dxf = super(DXFGraphic, self).load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, merged_ellipse_group_codes)
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_ellipse.name)
        # is_valid_ratio() takes floating point imprecision on the upper bound of
        # +/- 1.0 into account:
        assert is_valid_ratio(
            self.dxf.ratio
        ), f"axis-ratio out of range [{MIN_RATIO}, {MAX_RATIO}]"
        self.dxf.ratio = clamp_axis_ratio(self.dxf.ratio)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "center",
                "major_axis",
                "extrusion",
                "ratio",
                "start_param",
                "end_param",
            ],
        )

    @property
    def minor_axis(self) -> Vec3:
        dxf = self.dxf
        return ellipse.minor_axis(Vec3(dxf.major_axis), Vec3(dxf.extrusion), dxf.ratio)

    @property
    def start_point(self) -> Vec3:
        return list(self.vertices([self.dxf.start_param]))[0]

    @property
    def end_point(self) -> Vec3:
        return list(self.vertices([self.dxf.end_param]))[0]

    def construction_tool(self) -> ConstructionEllipse:
        """Returns construction tool :class:`ezdxf.math.ConstructionEllipse`."""
        dxf = self.dxf
        return ConstructionEllipse(
            dxf.center,
            dxf.major_axis,
            dxf.extrusion,
            dxf.ratio,
            dxf.start_param,
            dxf.end_param,
        )

    def apply_construction_tool(self, e: ConstructionEllipse) -> Ellipse:
        """Set ELLIPSE data from construction tool
        :class:`ezdxf.math.ConstructionEllipse`.

        """
        self.update_dxf_attribs(e.dxfattribs())
        return self  # floating interface

    def params(self, num: int) -> Iterable[float]:
        """Returns `num` params from start- to end param in counter-clockwise
        order.

        All params are normalized in the range [0, 2π).

        """
        start = self.dxf.start_param % math.tau
        end = self.dxf.end_param % math.tau
        yield from ellipse.get_params(start, end, num)

    def vertices(self, params: Iterable[float]) -> Iterable[Vec3]:
        """Yields vertices on ellipse for iterable `params` in WCS.

        Args:
            params: param values in the range from 0 to 2π in radians,
                param goes counter-clockwise around the extrusion vector,
                major_axis = local x-axis = 0 rad.

        """
        yield from self.construction_tool().vertices(params)

    def flattening(self, distance: float, segments: int = 8) -> Iterable[Vec3]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments, if the distance from the center
        of the approximation segment to the curve is bigger than `distance` the
        segment will be subdivided. Returns a closed polygon for a full ellipse
        where the start vertex has the same value as the end vertex.

        Args:
            distance: maximum distance from the projected curve point onto the
                segment chord.
            segments: minimum segment count

        """
        return self.construction_tool().flattening(distance, segments)

    def swap_axis(self):
        """Swap axis and adjust start- and end parameter."""
        e = self.construction_tool()
        e.swap_axis()
        self.update_dxf_attribs(e.dxfattribs())

    @classmethod
    def from_arc(cls, entity: DXFGraphic) -> Ellipse:
        """Create a new virtual ELLIPSE entity from an ARC or a CIRCLE entity.

        The new entity has no owner, no handle, is not stored in the entity database nor
        assigned to any layout!

        """
        assert entity.dxftype() in {"ARC", "CIRCLE"}, "ARC or CIRCLE entity required"
        attribs = entity.dxfattribs(drop={"owner", "handle", "thickness"})
        e = ellipse.ConstructionEllipse.from_arc(
            center=attribs.get("center", NULLVEC),
            extrusion=attribs.get("extrusion", Z_AXIS),
            # Remove all not ELLIPSE attributes:
            radius=attribs.pop("radius", 1.0),
            start_angle=attribs.pop("start_angle", 0),
            end_angle=attribs.pop("end_angle", 360),
        )
        attribs.update(e.dxfattribs())
        return cls.new(dxfattribs=attribs, doc=entity.doc)

    def transform(self, m: Matrix44) -> Ellipse:
        """Transform the ELLIPSE entity by transformation matrix `m` inplace."""
        e = self.construction_tool()
        e.transform(m)
        self.update_dxf_attribs(e.dxfattribs())
        self.post_transform(m)
        return self

    def translate(self, dx: float, dy: float, dz: float) -> Ellipse:
        """Optimized ELLIPSE translation about `dx` in x-axis, `dy` in y-axis
        and `dz` in z-axis, returns `self` (floating interface).

        """
        self.dxf.center = Vec3(dx, dy, dz) + self.dxf.center
        # Avoid Matrix44 instantiation if not required:
        if self.is_post_transform_required:
            self.post_transform(Matrix44.translate(dx, dy, dz))
        return self

    def to_spline(self, replace=True) -> Spline:
        """Convert ELLIPSE to a :class:`~ezdxf.entities.Spline` entity.

        Adds the new SPLINE entity to the entity database and to the
        same layout as the source entity.

        Args:
            replace: replace (delete) source entity by SPLINE entity if ``True``

        """
        from ezdxf.entities import Spline

        spline = Spline.from_arc(self)
        layout = self.get_layout()
        assert layout is not None, "valid layout required"
        if replace:
            replace_entity(self, spline, layout)
        else:
            add_entity(spline, layout)
        return spline

    def ocs(self) -> OCS:
        # WCS entity which supports the "extrusion" attribute in a
        # different way!
        return OCS()

    def audit(self, auditor: Auditor) -> None:
        if not self.is_alive:
            return
        super().audit(auditor)
        entity = str(self)
        major_axis = Vec3(self.dxf.major_axis)
        if major_axis.is_null:
            auditor.trash(self)
            auditor.fixed_error(
                code=AuditError.INVALID_MAJOR_AXIS,
                message=f"Removed {entity} with invalid major axis: (0, 0, 0).",
            )
            return
        axis_ratio = self.dxf.ratio
        if is_valid_ratio(axis_ratio):
            # remove possible floating point imprecision:
            self.dxf.ratio = clamp_axis_ratio(axis_ratio)
            return
        if abs(axis_ratio) > MAX_RATIO:
            self.swap_axis()
            auditor.fixed_error(
                code=AuditError.INVALID_ELLIPSE_RATIO,
                message=f"Fixed invalid axis-ratio in {entity} by swapping axis.",
            )
        elif abs(axis_ratio) < MIN_RATIO:
            self.dxf.ratio = clamp_axis_ratio(axis_ratio)
            auditor.fixed_error(
                code=AuditError.INVALID_ELLIPSE_RATIO,
                message=f"Fixed invalid axis-ratio in {entity}, set to {MIN_RATIO}.",
            )



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf
from ezdxf.audit import Auditor


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_new_group(doc):
    msp = doc.modelspace()
    group = doc.groups.new("test1")
    with group.edit_data() as g:
        g.append(msp.add_point((0, 0)))
        g.append(msp.add_line((1, 1), (2, 2)))
    assert len(group) == 2
    assert group[0].dxftype() == "POINT"
    assert group[1].dxftype() == "LINE"
    assert len(list(group.handles())) == 2
    handle = group[0].dxf.handle
    assert handle in group


def test_unique_groups(doc):
    doc.groups.new("test2")
    with pytest.raises(ValueError):
        doc.groups.new("test2")


def test_modify_group(doc):
    msp = doc.modelspace()
    group = doc.groups.new("test3")
    with group.edit_data() as g:
        g.append(msp.add_point((0, 0)))
        g.append(msp.add_line((1, 1), (2, 2)))
    assert len(group) == 2
    e = [
        msp.add_point((3, 3)),
        msp.add_point((4, 4)),
    ]
    group.extend(e)
    assert len(group) == 4


def test_can_not_add_invalid_block_entities(doc):
    group = doc.groups.new("test4")
    block = doc.blocks.new("Block4")
    point = block.add_point((0, 0))
    with pytest.raises(ezdxf.DXFStructureError):
        with group.edit_data() as g:
            g.append(point)


def test_can_not_add_invalid_table_entry(doc):
    group = doc.groups.new("test5")
    layer = doc.layers.get("0")
    with pytest.raises(ezdxf.DXFStructureError):
        with group.edit_data() as g:
            g.append(layer)


def test_audit_filters_invalid_entities(doc):
    group = doc.groups.new("test6")
    msp = doc.modelspace()
    block = doc.blocks.new("Block6")
    point1 = block.add_point((0, 0))  # invalid BLOCK entity
    point2 = msp.add_point((0, 0))  # valid model space entity ...
    point2.destroy()  # ... but destroyed
    layer = doc.layers.get("0")  # invalid table entry

    # hack, hack, hack - but since 2024-05-23 group.extend() validates the input data!
    group._data.extend([point1, point2, layer])
    auditor = Auditor(doc)
    group.audit(auditor)
    assert len(group) == 0


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import pathlib
import ezdxf
from ezdxf.entities import Insert
from ezdxf import dynblkhelper

BASEDIR = pathlib.Path(__file__)
DYN_BLOCKS = BASEDIR.parent / "data" / "dynblks.zip"


def test_direct_dynamic_block_references():
    doc = ezdxf.readzip(DYN_BLOCKS, "dynblk0.dxf")
    msp = doc.modelspace()

    references = msp.query("INSERT")
    assert len(references) == 4

    insert: Insert
    for insert in references:
        assert not insert.dxf.name.startswith("*"), "expected regular block reference"
        block = dynblkhelper.get_dynamic_block_definition(insert)
        assert block is not None
        assert block.name == insert.dxf.name


def test_indirect_dynamic_block_references():
    doc = ezdxf.readzip(DYN_BLOCKS, "dynblk1.dxf")
    msp = doc.modelspace()

    references = msp.query("INSERT")
    assert len(references) == 2

    insert: Insert
    for insert in references:
        assert insert.dxf.name.startswith("*"), "expected anonymous block reference"
        block = dynblkhelper.get_dynamic_block_definition(insert)
        assert block is not None
        assert block.name == "XYZ"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2024 Manfred Moitzi
# License: MIT License
import ezdxf
from ezdxf.math import Matrix44
import pytest

IDENTITY_MATRIX = list(list(Matrix44()))

MATRIX_CHECK = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new("R2007")
    return doc.modelspace()


def test_surface(msp):
    surface = msp.add_surface()
    assert surface.dxftype() == "SURFACE"
    # ACIS data is tested in test_body()
    assert len(surface.sat) == 0


def test_extruded_surface(msp):
    surface = msp.add_extruded_surface()
    assert surface.dxftype() == "EXTRUDEDSURFACE"
    assert len(surface.sat) == 0

    matrix = surface.transformation_matrix_extruded_entity
    assert list(matrix) == IDENTITY_MATRIX

    matrix = surface.sweep_entity_transformation_matrix
    assert list(matrix) == IDENTITY_MATRIX

    matrix = surface.path_entity_transformation_matrix
    assert list(matrix) == IDENTITY_MATRIX

    surface.transformation_matrix_extruded_entity = Matrix44(MATRIX_CHECK)
    assert list(surface.transformation_matrix_extruded_entity) == MATRIX_CHECK


def test_lofted_surface(msp):
    surface = msp.add_lofted_surface()
    assert surface.dxftype() == "LOFTEDSURFACE"
    assert len(surface.sat) == 0

    matrix = surface.transformation_matrix_lofted_entity
    assert list(matrix) == IDENTITY_MATRIX


def test_swept_surface(msp):
    surface = msp.add_swept_surface()
    assert surface.dxftype() == "SWEPTSURFACE"
    assert len(surface.sat) == 0


def test_revolved_surface(msp):
    surface = msp.add_revolved_surface()
    assert surface.dxftype() == "REVOLVEDSURFACE"
    assert len(surface.sat) == 0

    matrix = surface.transformation_matrix_revolved_entity
    assert list(matrix) == IDENTITY_MATRIX



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest
from ezdxf.entities import Hatch
from ezdxf.math import Vec2


@pytest.fixture
def all_edge_types_hatch():
    # used in test files:
    # 229c
    # 250
    hatch = Hatch.new(
        dxfattribs={
            "layer": "0",
            "color": "2",
            "elevation": (0.0, 0.0, 0.0),
            "extrusion": (0.0, 0.0, 1.0),
            "pattern_name": "SOLID",
            "solid_fill": 1,
            "associative": 0,
            "hatch_style": 0,
            "pattern_type": 1,
        },
    )
    # edge-path contains all supported edge types:
    ep = hatch.paths.add_edge_path(flags=1)
    ep.add_arc(  # clockwise oriented ARC
        center=(0.0, 13.0),
        radius=3.0,
        start_angle=-90.0,
        end_angle=90.0,
        ccw=False,
    )
    ep.add_ellipse(  # clockwise oriented ELLIPSE
        center=(0.0, 5.0),
        major_axis=(0.0, 5.0),
        ratio=0.6,
        start_angle=180.0,
        end_angle=360.0,
        ccw=False,
    )
    ep.add_line((0.0, 0.0), (10.0, 0.0))  # LINE
    ep.add_ellipse(  # counter-clockwise oriented ELLIPSE
        center=(10.0, 5.0),
        major_axis=(0.0, -5.0),
        ratio=0.6,
        start_angle=0.0,
        end_angle=180.0,
        ccw=True,
    )
    ep.add_arc(  # counter-clockwise oriented ARC
        center=(10.0, 13.0),
        radius=3.0,
        start_angle=270.0,
        end_angle=450.0,
        ccw=True,
    )
    ep.add_spline(  # SPLINE
        control_points=[
            Vec2(10.0, 16.0),
            Vec2(9.028174684192452, 16.0),
            Vec2(6.824943218065775, 12.14285714285714),
            Vec2(3.175056781934232, 19.85714285714287),
            Vec2(0.9718253158075516, 16.0),
            Vec2(0, 16.0),
        ],
        knot_values=[
            0.0,
            0.0,
            0.0,
            0.0,
            2.91547594742265,
            8.746427842267952,
            11.6619037896906,
            11.6619037896906,
            11.6619037896906,
            11.6619037896906,
        ],
        degree=3,
        periodic=0,
    )
    return hatch



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
# 1 plot unit (plu) = 0.025mm
# 40 plu = 1 mm
# 1016 plu = 1 inch
# 3.39 plu = 1 dot @300 dpi

from __future__ import annotations
from typing import Sequence
import math
from .deps import Vec2, NULLVEC2

INCH_TO_PLU = 1016
MM_TO_PLU = 40


class Page:
    def __init__(self, size_x: int, size_y: int):
        self.size_x = int(size_x)  # in plotter units (plu)
        self.size_y = int(size_y)  # plu

        self.p1 = NULLVEC2  # plu
        self.p2 = Vec2(size_x, size_y)
        self.user_scaling = False
        self.user_scale_x: float = 1.0
        self.user_scale_y: float = 1.0
        self.user_origin = NULLVEC2  # plu

    def set_scaling_points(self, p1: Vec2, p2: Vec2) -> None:
        self.reset_scaling()
        self.p1 = Vec2(p1)
        self.p2 = Vec2(p2)

    def apply_scaling_factors(self, sx: float, sy: float) -> None:
        self.set_ucs(self.user_origin, self.user_scale_x * sx, self.user_scale_y * sy)

    def set_scaling_points_relative_1(self, xp1: float, yp1: float) -> None:
        size = self.p2 - self.p1
        p1 = Vec2(self.size_x * xp1, self.size_y * yp1)
        self.set_scaling_points(p1, p1 + size)

    def set_scaling_points_relative_2(
        self, xp1: float, yp1: float, xp2: float, yp2: float
    ) -> None:
        p1 = Vec2(self.size_x * xp1, self.size_y * yp1)
        p2 = Vec2(self.size_x * xp2, self.size_y * yp2)
        self.set_scaling_points(p1, p2)

    def reset_scaling(self) -> None:
        self.p1 = NULLVEC2
        self.p2 = Vec2(self.size_x, self.size_y)
        self.set_ucs(NULLVEC2)

    def set_isotropic_scaling(
        self,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        left=0.5,
        bottom=0.5,
    ) -> None:
        size = self.p2 - self.p1
        delta_x = x_max - x_min
        delta_y = y_max - y_min
        scale_x = 1.0
        if abs(delta_x) > 1e-9:
            scale_x = size.x / delta_x
        scale_y = 1.0
        if abs(delta_y) > 1e-9:
            scale_y = size.y / delta_y

        scale = min(abs(scale_x), abs(scale_y))
        scale_x = math.copysign(scale, scale_x)
        scale_y = math.copysign(scale, scale_y)
        offset_x = (size.x - delta_x * scale_x) * left
        offset_y = (size.y - delta_y * scale_y) * bottom
        origin_x = self.p1.x + offset_x - x_min * scale_x
        origin_y = self.p1.y + offset_y - y_min * scale_y
        self.set_ucs(Vec2(origin_x, origin_y), scale_x, scale_y)

    def set_anisotropic_scaling(
        self, x_min: float, x_max: float, y_min: float, y_max: float
    ) -> None:
        size = self.p2 - self.p1
        delta_x = x_max - x_min
        delta_y = y_max - y_min
        scale_x = 1.0
        if abs(delta_x) > 1e-9:
            scale_x = size.x / delta_x
        scale_y = 1.0
        if abs(delta_y) > 1e-9:
            scale_y = size.y / delta_y
        origin_x = self.p1.x - x_min * scale_x
        origin_y = self.p1.y - y_min * scale_y
        self.set_ucs(Vec2(origin_x, origin_y), scale_x, scale_y)

    def set_ucs(self, origin: Vec2, sx: float = 1.0, sy: float = 1.0):
        self.user_origin = Vec2(origin)
        self.user_scale_x = float(sx)
        self.user_scale_y = float(sy)
        if abs(self.user_scale_x) < 1e-6:
            self.user_scale_x = 1.0
        if abs(self.user_scale_y) < 1e-6:
            self.user_scale_y = 1.0
        if math.isclose(self.user_scale_x, 1.0) and math.isclose(
            self.user_scale_y, 1.0
        ):
            self.user_scaling = False
        else:
            self.user_scaling = True

    def page_point(self, x: float, y: float) -> Vec2:
        """Returns the page location as page point in plotter units."""
        return self.page_vector(x, y) + self.user_origin

    def page_vector(self, x: float, y: float) -> Vec2:
        """Returns the user vector in page vector in plotter units."""
        if self.user_scaling:
            x = self.user_scale_x * x
            y = self.user_scale_y * y
        return Vec2(x, y)

    def page_points(self, points: Sequence[Vec2]) -> list[Vec2]:
        """Returns all user points as page points in plotter units."""
        return [self.page_point(p.x, p.y) for p in points]

    def page_vectors(self, vectors: Sequence[Vec2]) -> list[Vec2]:
        """Returns all user vectors as page vectors in plotter units."""
        return [self.page_vector(p.x, p.y) for p in vectors]

    def scale_length(self, length: float) -> tuple[float, float]:
        """Scale a length in user units to plotter units, scaling can be non-uniform."""
        return length * self.user_scale_x, length * self.user_scale_y
